{ Project : fwknopclient.dpr                                              }
 { Project : fwknopclient.dpr                                              }
 { Comment : fwknop is an SPA implimentation                               }
 {  Original License
#############################################################################
#
# File: fwknop
#
# Purpose: fwknop implements an authorization scheme known as Single Packet
#          Authorization (SPA) that requires only a single encrypted packet
#          to communicate various pieces of information including desired
#          access through an iptables policy and/or specific commands to
#          execute on the target system.  The main application of this
#          program is to protect services such as SSH with an additional
#          layer of security in order to make the exploitation of
#          vulnerabilities (both 0-day and unpatched code) much more
#          difficult.  For more information, see the fwknop(8) man page.
#
# Author: Michael Rash (mbr@cipherdyne.org)
#
# Copyright (C) 2004-2007 Michael Rash (mbr@cipherdyne.org)
#
# License (GNU Public License):
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
#    USA
#
#############################################################################
 }
 {    Date : 11/17/2007 8:03:00 PM                                         }
 {  Author : sgreven                                                       }
 {}

unit main;

interface

uses
  Classes, Controls, Forms, Graphics, Messages, SysUtils, Variants, Windows,
  Dialogs, IdBaseComponent, IdComponent, IdTCPConnection, StdCtrls,
  ComCtrls, IdDNSResolver, IdHTTP, IdUDPBase, IdUDPClient,
  DCPblockciphers, DCPcrypt2, DCPmd5, DCPrijndael, strutils,
  dateutils, DCPbase64, ExtCtrls, Menus, IdTCPClient;

type
  TForm1 = class( TForm)
    Edit4           : TEdit;
    Button2         : TButton;
    GroupBox1       : TGroupBox;
    rbPCAP          : TRadioButton;
    enc_pcap_port   : TEdit;
    Label4          : TLabel;
    StatusBar1      : TStatusBar;
    edtClientKey    : TEdit;
    GroupBox2       : TGroupBox;
    comboProto      : TComboBox;
    edtprotoport    : TEdit;
    MDebug          : TMemo;
    Label6          : TLabel;
    edtUsername     : TEdit;
    Label7          : TLabel;
    GroupBox3       : TGroupBox;
    rbYes           : TRadioButton;
    rbNo            : TRadioButton;
    mm1             : TMainMenu;
    Settings1       : TMenuItem;
    Debug1          : TMenuItem;
    procedure Button2Click ( Sender: TObject);
    procedure Debug1Click ( Sender: TObject);
  private
    function padr ( encryptstr: string): string;

    { Private declarations }
  public
    { Public declarations }
  end;

type
  Saltret = array[0..2] of string;

var
  Form1: TForm1;
(*################################################
#  Some constants have to be defined here to be used later
#  in the code.
#  Keysize   : defines the keysize used in Rijndael (Bytes)
#              32 Bytes = 256 bits
#  Blocksize : defines the Blocksize used in Rijndael (Bytes)
#              16 Bytes is 128 bits
#  Version   : is used to define the fwknop client version (
#              This is included in the SPA packet)
################################################*)
const
  Version   = '1.8.2';
  KeySize   = 32;
  BlockSize = 16;

implementation

{$R *.dfm}
{$D-,L-,O+,Q-,R-,Y-,S-}

(*################################################
#  Function    : RandInt
#  Arguments   : NumberOfDigits
#  Returns     : Semi Random string of digits (16 in this case), used as the
#                first 32 bytes of the SPA packet, this introduces
#                randomness into the cipher, and prevents brute
#                force on a large subset of SPA packets.
#   Return Type: String;
################################################*)

function RandInt ( NumberOfDigits : Integer ): string;
var
  i, j:   integer;
  varstr: string;
begin
  for j := 1 to NumberOfDigits do
  begin
    Randomize;
    i      := Random ( 10);
    varstr := varstr + IntToStr ( i);
  end;
  Result := varstr;
end;

(*################################################
#  Function    : StringToHex
#  Arguments   : Data
#  Returns     : Hex Byte
#  Return Type : String;
################################################*)

function StringToHex(Data: string): string;
var
  i, i2: Integer;
  s: string;
begin
  i2 := 1;
  for i := 1 to Length(Data) do
  begin
    Inc(i2);
    if i2 = 2 then
    begin
      s  := s + ' ';
      i2 := 1;
    end;
    s := s + IntToHex(Ord(Data[i]), 2);
  end;
  Result := s;
end;

(*################################################
#  Function    : GenSalt
#  Arguments   : MaxLen,Restrictions
#  Returns     : Returns Random Hex string
#              Possible Restrictions:
#              1 - Any characters
#              2 - Letters and Numbers
#              3 - Numbers
#              4 - Letters
#   Return Type: String;
################################################*)

Function GenSalt(MaxLen,Restrictions:Integer):String;
Var St:Byte;
    i : Integer;
    Total:String;
begin
    i := 0;
    Repeat
    Randomize;
    St := Random(255);
    If Restrictions = 1 then
    begin
      Total := Total + StringToHex(chr(St));
      i := i+1;
    end;
    If Restrictions = 2 then
    begin
      If (St >= 48) and (St <= 57) or (St >= 65)
              and (St <= 90) or (St >= 97) and (St <= 122) then
      begin
        Total := Total + StringToHex(chr(St));
        i := i+1;
      end;
    end;
    If Restrictions = 3 then
    begin
      If (St >= 48) and (St <= 57) then
      begin
        Total := Total + StringToHex(chr(St));
        i := i + 1;
      end;
    end;
    If Restrictions = 4 then
    begin
      If (St >= 65) and (St <= 90) or (St >= 97) and (St <= 122) then
      begin
        Total := Total + StringToHex(chr(St));
        i := i + 1;
      end;
    end;
    Until i = MaxLen;
      Result := Total;
end;

(*################################################
#  Function   : TimeStamp
#  Arguments  : None
#  Returns    : Current Time as a unix timestamp (seconds
#               since the epoch 1 Jan 1970). This is returned
#               as GMT to be compatible with perls localtime().
#  Return Type: String;
################################################*)

function Timestamp ( ): string;
var
  tstamp: integer;
  secondsoffset: integer;
  timezoneinfo: TTimezoneinformation;
begin
  GetTimezoneInformation ( timezoneinfo);
  secondsoffset := timezoneinfo.Bias * 60;
  tstamp := DateTimeToUnix ( Now) + secondsoffset;
  Result := IntToStr ( tstamp);
end;

(*################################################
#  Function   : HexStrToStr
#  Arguments  : TheEx
#  Returns    : This function returns the String equivalent of
#               the Hex string passed to it, used to convert
#               the salt to HEX. Also used in the recreation of
#               the Crypt::CBC function _salted_key_and_iv(), it
#               is also used in the DoEncryption function.  The
#               Hex atring is passed with each byte - space separated
#  Return Type: String;
################################################*)

function HexStrToStr ( TheEx: string): string;
var
  TheList: TStringList;
  n: integer;
begin
  Result  := '';
  TheList := TStringList.Create;
  try
    TheList.DelimitedText := TheEx;
    for n := 0 to Pred ( TheList.Count) do
    begin
      Result := Result + Chr ( StrToInt ( '$' + TheList [n]));
    end;
  finally
    TheList.Free;
  end;
end;

(*################################################
#  Function   : pad_block
#  Arguments  : S [which is the string to pad]
#  Returns    : This function returns a string pkcs #5 padded
#               string, which complies with the RSA std
#               6.2 Encryption-block formatting
#
#                The message M and a padding string PS shall be formatted
#                into an octet string EB, the encryption block.
#
#                       EB = M || PS .                    (1)
#
#                The padding string PS shall consist of 8 - (||M|| mod 8)
#                octets all having value 8 - (||M|| mod 8). (This makes the
#                length of the encryption block EB a multiple of eight
#                octets.) In other words, the encryption block EB satisfies
#                one of the following statements:
#
#                EB = M || 01 -- if ||M|| mod 8 = 7 ;
#                EB = M || 02 02 -- if ||M|| mod 8 = 6 ;
#                              .
#                              .
#                              .
#                EB = M || 08 08 08 08 08 08 08 08 -- if ||M|| mod 8 = 0 .
#
#  Return Type: String;
################################################*)

function PadBlock ( s: string): string;
var
  clear_txt: string;
  i, btlen, btmultiplier: integer;
begin
  clear_txt := s;
  btmultiplier := length ( clear_txt) div blocksize;
  btlen := ( ( btmultiplier + 1) * blocksize) - length ( clear_txt);
  i     := 1;
  while i <= btlen do
  begin
    clear_txt := clear_txt + chr ( Ord ( btlen));
    i := i + 1;
  end;
  Result := AnsiMidStr ( clear_txt, 0, length ( clear_txt));
end;

(*################################################
#  Function   : PadPass
#  Arguments  : S [which is the password to pad]
#  Returns    : This function returns the password passed to it
#               padded to the blocksize, ie. It returns the
#               password padded to 16 Bytes. The padding character is
#               '0', this is the equivalent code to the perl client
#                   while (length($clear_txt) < $enc_blocksize-1)
#                      $clear_txt .= chr(0);
#                   }
#  Return Type: String;
################################################*)

function PadPass ( s: string): string;
var
  newpass: string;
begin
  newpass := s;
  while length ( newpass) < blocksize do
  begin
    newpass := newpass + '0';
  end;
  Result := newpass;
end;

(*################################################
#  Function   : MD5
#  Arguments  : Password
#  Returns    : This is a wrapper to the MD5 hashing component
#               It returns a 16 Byte MD5 hash (RFC1321).
#               This is used in the _salted_key_and_iv function, since
#               openssl uses MD5 instead of sha to do key derivation.
#  Return Type: String;
################################################*)

function MD5 ( Password: string): string;
const
  HASHSIZE_BYTES = 16;
var
  Digest: array[0..HASHSIZE_BYTES - 1] of byte;
  i:      integer;
begin
  with TDCP_md5.Create ( nil) do
  begin
    try
      Init ( );
      UpdateStr ( Password);
      final ( Digest [0]);
      Result := '';
      for i := 0 to HASHSIZE_BYTES - 1 do
      begin
        if i <> 0 then
        begin
          Result := Result + ' ';
        end;
        Result := Result + IntToHex ( ( Digest [i]), 2);
      end;
      Result := LowerCase ( Result);
    finally
      Free ( );
    end;
  end;
end;

(*################################################
#  Function   : _salted_key_and_iv
#  Arguments  : Password
#  Returns    : This is most likely the most frustrating function
#               experienced in this whole project.  This is a function
#               to derive an IV and a key to be used in the Rijndael encryption
#               process.  The 'Password' refered to in Crypt::CBC is actually
#               a passphrase, and this function derives a password and an
#               Initialization Vector from 8 bytes of Salt (which can be random).
#               This is probably the nexus of the issues experienced while coding
#               this project.  Thanks to Michael Rash for modifying the fwknopd code
#               to do a Hex Dump in debug mode, which helped determine what was wrong.
#               Documentation on the innards of OpenSSL's Rijndael doesn't help too
#               much here, EVP_BytesToKey() is the equivalent OpenSSL function.
#               OpenSSL documentation follows:
#               SYNOPSIS
#                #include <openssl/evp.h>
#
#                int EVP_BytesToKey(const EVP_CIPHER *type,const EVP_MD *md,
#                       const unsigned char *salt,
#                       const unsigned char *data, int datal, int count,
#                       unsigned char *key,unsigned char *iv);
#
#               EVP_BytesToKey() derives a key and IV from various parameters.
#               type is the cipher to derive the key and IV for. md is the message digest to use.
#               The salt paramter is used as a salt in the derivation: it should
#               point to an 8 byte buffer or NULL if no salt is used. data is a buffer containing
#               datal bytes which is used to derive the keying data. count is the iteration count to use.
#               The derived key and IV will be written to key and iv respectively.
#
#               The key and IV is derived by concatenating D_1, D_2, etc until enough data is available
#               for the key and IV. D_i is defined as:
#
#                  D_i = HASH^count(D_(i-1) || data || salt)
#
#               where || denotes concatentaion, D_0 is empty, HASH is the digest algorithm in use,
#               HASH^1(data) is simply HASH(data), HASH^2(data) is HASH(HASH(data))
#               and so on.
#
#               The initial bytes are used for the key and the subsequent bytes
#               for the IV.
#  Return Type: String;
################################################*)

function _salted_key_and_iv ( Pass, Salt: string): Saltret;
var
  saltlen: integer;
  Data, d: string;
begin
  saltlen := length ( Salt);
  if not saltlen = 8 then
  begin
    raise Exception.Create ( 'Salt must be 8 bytes long');
  end;

  Data := '';
  d    := '';

  d    := md5 ( pass + salt);
  Data := HexStrToStr ( d);
  d    := md5 ( Data + pass + salt);
  Data := Data + HexStrToStr ( d);
  Result[0] := Data;
  d    := md5 ( HexStrToStr ( d) + pass + salt);
  Result[1] := HexStrToStr ( d);
end;

(*################################################
#  Function   : DoEncryption
#  Arguments  : InputString, MyPassword
#  Returns    : Returns the base64 encoded packet to be sent to the fwknop server
#               Once again thanks to Michael Rash for his Hex Dump function as well as
#               to hours and hours of reading Crypt::CBC and wireshark, to reverse
#               engineer the Salt header to the packet.
#               Suffice to say, the First part of the header is the string
#               Salted__ followed by the salt 8 Bytes.  The salt is used in Crypt::CBC to
#               derive the key + the password, which is used in the decryption process.
#               Another thing to remember here is that the payload is padded using the
#               PadBlock function which pads the final block to make it a full size block
#               using pkcs #5 padding.  This means that the final block is padded with the
#               hex equivalent of the number of padding blocks.
#  Return Type: String;
################################################*)

function DoEncryption ( InputString, MyPassword: string)
: string;
var
  Key_iv: saltret;
  Slt, encrypted, Key, iv, cleardata, packet: string;
  DCP_rijndaelEnc: TDCP_rijndael;
begin
  DCP_rijndaelEnc := nil;
  try
    DCP_rijndaelEnc := TDCP_rijndael.Create ( nil);
//    slt    := HexStrToStr ( '69 7c 14 8a c8 27 61 40');
    slt := HexStrToStr (GenSalt(8,1));
    cleardata := PadBlock ( InputString);
    Key_iv := _salted_key_and_iv ( PadPass ( MyPassword), slt);
    Key    := Key_iv [0];
    iv     := Key_iv [1];
    DCP_rijndaelEnc.Init ( Key [1], 256, @iv [1]);
    SetLength ( encrypted, length ( cleardata));
    DCP_rijndaelEnc.EncryptCBC ( cleardata [1], encrypted [1],
      length ( cleardata));
    DCP_rijndaelEnc.Reset;
    packet := 'Salted__' + slt + encrypted;
    Result := AnsiMidStr ( Base64EncodeStr ( packet), 0,
      length ( Base64EncodeStr ( packet)) - 2);
  finally
    DCP_rijndaelEnc.Free;
  end;
end;

(*################################################
#  Function   : PadR
#  Arguments  : encryptstr
#  Returns    : Returns the passphrase padded with '0' to 16 chars,
#               for compatability with fwknop
#                ### pad out to 16 chars
#                while (length($enc_key) < $enc_blocksize) {
#                   $enc_key .= '0';
#                }
#  Return Type: String;
################################################*)

function TForm1.PadR ( encryptstr: string): string;
var
  j:      integer;
  varstr: string;
begin
  varstr := encryptstr;
  for j := length ( encryptstr) to 15 do
  begin
    varstr := varstr + '0';
  end;
  Result := varstr;
end;

procedure TForm1.Button2Click ( Sender: TObject);
var
  ipaddress, cleartext, randomstr, tstamp, premd5, postmd5,
  md5str, encrypted, key, md5char: string;
  i:      integer;
  Hash:   TDCP_md5;
  IdUDPClient1 : TIdUDPClient;
  IdHTTP1 : TIdHTTP;
  Digest: array[0..15] of byte;
begin
  randomstr := RandInt(16);
  tstamp    := Timestamp;
  if length ( edtClientKey.Text) <= 7 then
  begin
    StatusBar1.Panels[1].Text :=
      'Client Key must be longer than 8 chars';
      raise Exception.Create('Client Key must be longer than 8 chars');
  end
  else
  begin
    StatusBar1.Panels[1].Text := 'Client Key OK';
  end;
  if rbYes.Checked then
  begin
    IdHTTP1 := Nil;
    try
      IdHTTP1 := TIdHTTP.Create(Nil);
      screen.Cursor := crHourGlass;
      try
        IdHTTP1.Port := 80;
        ipaddress := idhttp1.Get ( 'http://www.whatismyip.org');
        StatusBar1.Panels[1].Text :=
          'Sucessfully Resolved External IP';
      except
        on E: Exception do
          StatusBar1.Panels[1].Text := E.Message;
      end;
    finally
      IdHTTP1.Free;
      screen.Cursor := crDefault;
    end;
  end
  else
  begin
    ipaddress := '0.0.0.0';
  end;
  MDebug.Clear;
  cleartext := randomstr + ':' + edtUsername.Text + ':' + tstamp +
    ':' + Version + ':' + '1' + ':' + ipaddress + ','
    + comboProto.Text + '/' + edtprotoport.Text;
  premd5    := randomstr + ':' + Base64EncodeStr ( edtUsername.Text) +
    ':' + tstamp + ':' + Version + ':' + '1' + ':' +
    Base64EncodeStr ( ipaddress + ',' + comboProto.Text +
    '/' + edtprotoport.Text);
  Hash      := TDCP_md5.Create ( nil);
  Hash.Init;
  Hash.UpdateStr ( premd5);
  Hash.Final ( Digest);
  for i := 0 to ( Sizeof ( Digest) - 1) do
  begin
    md5str  := md5str + chr ( Digest [i]);   // convert Digest to a
    md5char := md5char + chr ( Digest [i]);  // base64
  end;
  Hash.Free;
  md5str  := Base64EncodeStr ( md5str);
  md5str  := StringReplace ( md5str, '==', '',
    [rfReplaceAll, rfIgnoreCase]);
  postmd5 := premd5 + ':' + md5str;
  key     := PadR ( edtClientKey.Text);
  encrypted := DoEncryption ( postmd5, edtClientKey.Text);

  if Debug1.Checked then
  begin
    MDebug.Clear;
    MDebug.Lines.Add ( '[+] ** Running in client debug mode. **');
    MDebug.Lines.Add ( '[+] Starting fwknop client.');
    MDebug.Lines.Add ( '');
    MDebug.Lines.Add ( '[+] Building encrypted single-packet authorization (SPA) message...');
    MDebug.Lines.Add ( '[+] Packet fields:');
    MDebug.Lines.Add ( '');
    MDebug.Lines.Add ( '');
    MDebug.Lines.Add ( '        Random data: ' + randomstr);
    MDebug.Lines.Add ( '        Username: ' + edtUsername.Text);
    MDebug.Lines.Add ( '        Timestamp ' + tstamp);
    MDebug.Lines.Add ( '        Version: ' + Version);
    MDebug.Lines.Add ( '        Action: 1 (access mode)');
    MDebug.Lines.Add ( '        Access: ' + ipaddress + ',' + comboProto.Text +
      '/' + edtprotoport.Text);
    MDebug.Lines.Add ( '        MD5 sum: ' + md5str);
    MDebug.Lines.Add ( '');
    MDebug.Lines.Add ( '[+] Clear text message (not base64 encoded):');
    MDebug.Lines.Add ( cleartext);
    MDebug.Lines.Add ( '[+] Clear text message (base64 encoded):');
    MDebug.Lines.Add ( premd5);
    MDebug.Lines.Add ( '');
    MDebug.Lines.Add ( '[+] Encrypted message:');
    MDebug.Lines.Add ( encrypted);
  end;
  IdUDPClient1 := Nil;
  try
    IdUDPClient1 := TIdUDPClient.Create(Nil);
    IdUDPClient1.BufferSize := 8192;
    IdUDPClient1.Host := Edit4.Text;
    IdUDPClient1.Port := StrToInt ( enc_pcap_port.Text);
    IdUDPClient1.Send ( encrypted);
    if Debug1.Checked then
    begin
      MDebug.Lines.Add ( '');
      MDebug.Lines.Add ( '[+] Sending ' + IntToStr (
              length ( encrypted)) + ' Byte message to ' +
              Edit4.Text + ' over udp/' + enc_pcap_port.Text + '...');
    end;
  finally
    IdUDPClient1.Free;
  end;
end;

procedure TForm1.Debug1Click ( Sender: TObject);
begin
  if Debug1.Checked then
  begin
    MDebug.Show;
  end
  else
  begin
    MDebug.Hide;
  end;
end;

end.

